@@ -1,3 +1,13 @@
+get_worker='''
+query GetWorker($code: String!){
+  worker(findBy:{ code: $code}){
+    id
+    code
+    name
+  }
+}
+'''
+
 all_jobs='''
 query WorkerJobs($worker: String){
     jobs(
