@@ -3,6 +3,7 @@ import time
 import json
 from json.decoder import JSONDecodeError
 import multiprocessing as mp
+from multiprocessing import SimpleQueue
 from datetime import datetime, timedelta
 from enum import Enum
 from pygqlc import GraphQLClient
@@ -55,15 +56,23 @@ class ValiotWorker():
     self.lockedJobs = []
     self.gql = GraphQLClient()
     self.worker = None
+    self.worker_code = None
     self.pollingMode = PollingMode.QUERY
     self.context = None
     self.unsubJobCreated = None
+    # logging attributes
+    self.stopLogging = False
+    self.loggingProcess = None
+    self.log = None
+    self.logQueues = {
+        'main': SimpleQueue()  # queue for the main thread (Orchestator)
+    }
   
   def setClient(self, client):
     self.gql = client
   
   def setWorker(self, worker):
-    self.worker = worker
+    self.worker_code = worker
   
   def setPollingMode(self, mode):
     self.pollingMode = mode
@@ -110,7 +119,7 @@ class ValiotWorker():
 
   def not_found(self, job_id, update_job, kwargs):
     from .Notifications import nextNotifOld
-    log(LogLevel.ERROR, 'Function not found :(')
+    self.log(LogLevel.ERROR, 'Function not found :(')
     message= 'Funcion no habilitada, favor de contactar a administrador'
     notificationData = {}
     if(nextNotifOld(kwargs['job'])): 
@@ -163,7 +172,7 @@ class ValiotWorker():
   def lockQueue(self, queue_name):
     variables = {
       'queue': queue_name,
-      'worker': self.worker,
+      'worker': self.worker['code'],
       'active': True,
       'startAt': getUtcDate(datetime.now()).strftime("%Y-%m-%dT%H:%M:%SZ")
     }
@@ -192,7 +201,7 @@ class ValiotWorker():
       lock = self.lockQueue(job["queue"]["name"])
     # Obtiene job
     fn_name, job_dict = self.getJobData(job)
-    log(LogLevel.INFO, f"running {fn_name}")
+    self.log(LogLevel.INFO, f"running {fn_name}")
     job_fn = job_dict['function']
     # Obtiene inputs
     if job['input']:
@@ -209,7 +218,7 @@ class ValiotWorker():
           'output': f'Job input error: {str(e)}'
         }
         uploaders.update_job(variables)
-        log(LogLevel.ERROR, e)
+        self.log(LogLevel.ERROR, e)
         return
     else:
       kwargs = job_dict
@@ -232,6 +241,7 @@ class ValiotWorker():
     kwargs['update_job'] = update_this_job  # couple of aliases for the same fn
     # acceso directo a atributos del queue al que pertenece el job
     kwargs['queue'] = job['queue']
+    kwargs['log'] = self.getProcessLoggerFunction(queue=job['queue'], job=job)
     kwargs['kwargs'] = {**kwargs}
     variables = {
       'id': int(job['id']),
@@ -247,15 +257,15 @@ class ValiotWorker():
         try:
           job_fn(**kwargs)
         except TypeError as e:
-          log(LogLevel.ERROR,
+          self.log(LogLevel.ERROR,
               f'Error starting job from queue {fn_name}:\n{str(e)}')
-          log(LogLevel.ERROR,
+          self.log(LogLevel.ERROR,
               f'Did you forget to ignore additional params in the job definition?:')
-          log(LogLevel.ERROR, f'@vw.job(...)')
-          log(LogLevel.ERROR, f'def my_job(my_param, my_param2, **_) # <== can be **kwargs or **_')
-          log(LogLevel.ERROR, f'\t...')
+          self.log(LogLevel.ERROR, f'@vw.job(...)')
+          self.log(LogLevel.ERROR, f'def my_job(my_param, my_param2, **_) # <== can be **kwargs or **_')
+          self.log(LogLevel.ERROR, f'\t...')
       p = self.context.Process(target=safe_fn, kwargs=(kwargs))
-      log(LogLevel.INFO, f'started job with ID {job["id"]}')
+      self.log(LogLevel.INFO, f'started job with ID {job["id"]}')
       p.start()
       if lock:
         self.lockedJobs.append({
@@ -264,7 +274,7 @@ class ValiotWorker():
           'lock': lock
         })
     else:
-      log(LogLevel.ERROR, f'Error starting job (queue={fn_name})')
+      self.log(LogLevel.ERROR, f'Error starting job (queue={fn_name})')
   
   def queue_locked(self, queue_name):
     queues_by_name = __.group_by(self.currentQueues, 'name')
@@ -274,39 +284,39 @@ class ValiotWorker():
     return __.get(current_queue, 'locks.0.active') # locked if last lock is active
 
   def processJobs(self):
-    jobs, errors = self.gql.query(queries.all_jobs, variables={'worker': self.worker})
+    jobs, errors = self.gql.query(queries.all_jobs, variables={'worker': self.worker['code']})
     if errors:
       message = errors[0]
-      log(LogLevel.ERROR, f'Error! msg: {message}')
+      self.log(LogLevel.ERROR, f'Error! msg: {message}')
       return
     if jobs is not None and len(jobs) > 0:
-      log(LogLevel.WARNING, f"Found {len(jobs)} jobs")
+      self.log(LogLevel.WARNING, f"Found {len(jobs)} jobs")
       for job in jobs:
         if self.queue_locked(job['queue']['name']):
-          log(LogLevel.ERROR, f"{job['queue']['name']} is LOCKED, job aborted")
+          self.log(LogLevel.ERROR, f"{job['queue']['name']} is LOCKED, job aborted")
           self.abortJob(job)
         elif job['queue']['enabled']:
           self.runJob(job)
         else:
-          log(LogLevel.WARNING, f"{job['queue']['name']} is DISABLED, skipping job")
+          self.log(LogLevel.WARNING, f"{job['queue']['name']} is DISABLED, skipping job")
     else:
-      log(LogLevel.INFO, "Listening to events...")
+      self.log(LogLevel.INFO, "Listening to events...")
 
   def abortStaleJobs(self):
-    jobs, errors = self.gql.query(queries.waiting_jobs, variables={'worker': self.worker})
+    jobs, errors = self.gql.query(queries.waiting_jobs, variables={'worker': self.worker['code']})
     if (errors):
-      log(LogLevel.ERROR, f'errors: {errors}')
+      self.log(LogLevel.ERROR, f'errors: {errors}')
     elif (len(jobs) > 0):
-      log(LogLevel.ERROR, f'Deleting {len(jobs)} stale jobs:')
+      self.log(LogLevel.ERROR, f'Deleting {len(jobs)} stale jobs:')
       for job in jobs:
         result, errors = self.gql.mutate(mutations.delete_job, {'id': job['id']})
         if (not errors):
-          log(LogLevel.ERROR, f'deleted stale job {job["id"]}:')
+          self.log(LogLevel.ERROR, f'deleted stale job {job["id"]}:')
         else:
-          log(LogLevel.ERROR, f'errors deleting stale job: {errors}')
+          self.log(LogLevel.ERROR, f'errors deleting stale job: {errors}')
 
   def updateAvailableQueues(self):
-    log(LogLevel.SUCCESS, 'Actualizando Jobs disponibles')
+    self.log(LogLevel.SUCCESS, 'Actualizando Jobs disponibles')
     for name, queue in self.queues.items():
       enabled = __.get(queue, 'enabled', None)
       description = __.get(queue, 'description', None)
@@ -334,20 +344,20 @@ class ValiotWorker():
         variables=variables
       )
       if (errors):
-        log(LogLevel.ERROR, f'Error actualizando Queue {name}.')
-        log(LogLevel.ERROR, errors)
+        self.log(LogLevel.ERROR, f'Error actualizando Queue {name}.')
+        self.log(LogLevel.ERROR, errors)
       else:
-        log(LogLevel.SUCCESS, f'Queue {name} actualizado correctamente')
+        self.log(LogLevel.SUCCESS, f'Queue {name} actualizado correctamente')
 
   def unlockFinishedJobs(self, force=False):
     unlocked = []
     for i, job_lock in enumerate(self.lockedJobs):
       if not job_lock['process'].is_alive() or force:
         if self.unlockQueue(job_lock['job']['queue']['name'], job_lock['lock']):
-          log(LogLevel.SUCCESS, f'{job_lock["job"]["queue"]["name"]} UNLOCKED')
+          self.log(LogLevel.SUCCESS, f'{job_lock["job"]["queue"]["name"]} UNLOCKED')
           unlocked.append(i)
         else:
-          log(LogLevel.ERROR, f'error unlocking {job_lock["job"]["queue"]["name"]}')
+          self.log(LogLevel.ERROR, f'error unlocking {job_lock["job"]["queue"]["name"]}')
     for u in unlocked:
       self.lockedJobs.pop(u)
 
@@ -358,7 +368,7 @@ class ValiotWorker():
         continue  # nothing to run
       elapsed = current_time - queue['start_time']
       if elapsed >= queue['config']['timeout']:
-        log(LogLevel.WARNING, f'Event {queue["name"]} timed out, running...')
+        self.log(LogLevel.WARNING, f'Event {queue["name"]} timed out, running...')
         self.runEvent(queue['name'])
   
   def runFrequencyQueues(self):
@@ -366,7 +376,7 @@ class ValiotWorker():
     utc = timezone("UTC")
     queuesByType = __.group_by(self.currentQueues, 'type')
     freqQueues = __.get(queuesByType, 'FREQUENCY', default=[])
-    # log(LogLevel.WARNING, f'checking {len(freqQueues)} frequency queues')
+    # self.log(LogLevel.WARNING, f'checking {len(freqQueues)} frequency queues')
     for queue in freqQueues:
       # if it has never been ran, set the lastRunAt as a very old date (forces next run)
       lastRunAt = __.get(queue, 'lastRunAt', None)
@@ -388,7 +398,7 @@ class ValiotWorker():
             mutation=mutations.create_job,
             variables={
               'queueName': queue['name'],
-              'worker': self.worker
+              'worker': self.worker['code']
             }
           )
   
@@ -402,30 +412,30 @@ class ValiotWorker():
       if not job:
         return # Possibly a job for another worker
       if self.queue_locked(job['queue']['name']):
-        log(LogLevel.ERROR, f"{job['queue']['name']} is LOCKED, job aborted")
+        self.log(LogLevel.ERROR, f"{job['queue']['name']} is LOCKED, job aborted")
         self.abortJob(job)
       elif job['queue']['enabled']:
         self.runJob(job)
       else:
-        log(LogLevel.WARNING, f"{job['queue']['name']} is DISABLED, skipping job")
+        self.log(LogLevel.WARNING, f"{job['queue']['name']} is DISABLED, skipping job")
 
     self.unsubJobCreated = self.gql.subscribe(
       query=subscriptions.JOB_CREATED,
-      variables={'workerCode': self.worker},
+      variables={'workerCode': self.worker['code']},
       callback=runOnDemandJob)
     if (self.unsubJobCreated):
-      log(LogLevel.SUCCESS, 'SUBSCRIBED TO JOB CREATIONS')
+      self.log(LogLevel.SUCCESS, 'SUBSCRIBED TO JOB CREATIONS')
     else:
-      log(LogLevel.ERROR, 'ERROR SUBSCRIBING TO JOB CREATIONS')
+      self.log(LogLevel.ERROR, 'ERROR SUBSCRIBING TO JOB CREATIONS')
   
   def runEvent(self, queue_name):
     queuesByName = __.group_by(self.currentQueues, 'name')
     currentQueue = __.get(queuesByName, f'{queue_name}.0')
     if self.queue_locked(currentQueue['name']):
-      log(LogLevel.WARNING, f"{currentQueue['name']} is LOCKED, event postponed")
+      self.log(LogLevel.WARNING, f"{currentQueue['name']} is LOCKED, event postponed")
       return
     elif not currentQueue['enabled']:
-      log(LogLevel.WARNING, f'Event triggered for {queue_name} but Queue not enabled, skipping...')
+      self.log(LogLevel.WARNING, f'Event triggered for {queue_name} but Queue not enabled, skipping...')
       return
     data = self.eventQueues[queue_name]["mailbox"]
     # ! try to run job
@@ -438,12 +448,12 @@ class ValiotWorker():
         mutation=mutations.run_event_job,
         variables={
             'queueName': queue_name,
-            'worker': self.worker,
+            'worker': self.worker['code'],
             'lastRunAt': getUtcDate(datetime.now()).strftime("%Y-%m-%dT%H:%M:%SZ")
         }
     )
     if errors:
-      log(LogLevel.ERROR, f'Error starting event:\n{errors}')
+      self.log(LogLevel.ERROR, f'Error starting event:\n{errors}')
       return
     job = __.get(response, 'job.result')
     def update_this_job(variables):
@@ -463,18 +473,19 @@ class ValiotWorker():
       'job': {**job, 'update': update_this_job},
       'update_job': update_this_job,
       'queue': currentQueue,
-      'data': data
+      'data': data,
+      'log': self.getProcessLoggerFunction(queue=currentQueue, job=job)
     }
     kwargs['kwargs'] = {**kwargs}
     def safe_fn(**kwargs):
       try:
         self.eventQueues[queue_name]['function'](**kwargs)
       except TypeError as e:
-        log(LogLevel.ERROR, f'Error starting job from queue {queue_name}:\n{str(e)}')
-        log(LogLevel.ERROR, f'Did you forget to ignore additional params in the job definition?:')
-        log(LogLevel.ERROR, f'@vw.job(...)')
-        log(LogLevel.ERROR, f'def my_job(my_param, my_param2, **_) # <== can be **kwargs or **_')
-        log(LogLevel.ERROR, f'\t...')
+        self.log(LogLevel.ERROR, f'Error starting job from queue {queue_name}:\n{str(e)}')
+        self.log(LogLevel.ERROR, f'Did you forget to ignore additional params in the job definition?:')
+        self.log(LogLevel.ERROR, f'@vw.job(...)')
+        self.log(LogLevel.ERROR, f'def my_job(my_param, my_param2, **_) # <== can be **kwargs or **_')
+        self.log(LogLevel.ERROR, f'\t...')
     p = self.context.Process( target=safe_fn, kwargs=(kwargs))
     p.start()
     if lock:
@@ -502,7 +513,7 @@ class ValiotWorker():
             Data received from the event
         """
         if len(self.eventQueues[queue["name"]]["mailbox"]) == 0:
-          log(LogLevel.WARNING, f'EVENT { queue["name"] } TRIGGERED, BUILDING BATCH')
+          self.log(LogLevel.WARNING, f'EVENT { queue["name"] } TRIGGERED, BUILDING BATCH')
           self.eventQueues[queue["name"]]["start_time"] = time.time()
         self.eventQueues[queue["name"]]["mailbox"].append(data)
         if len(self.eventQueues[queue["name"]]["mailbox"]) >= self.eventQueues[queue["name"]]['config']['max_batch']:
@@ -510,15 +521,67 @@ class ValiotWorker():
           self.runEvent(queue["name"])
       unsub = self.gql.subscribe(query=queue['query'], callback=onEventTriggered)
       if (unsub):
-        log(LogLevel.SUCCESS, f'REGISTERED LISTENER FOR EVENT QUEUE { queue["name"] }')
+        self.log(LogLevel.SUCCESS, f'REGISTERED LISTENER FOR EVENT QUEUE { queue["name"] }')
       else:
-        log(LogLevel.ERROR, f'ERROR REGISTERING LISTENER FOR EVENT QUEUE { queue["name"] }')
+        self.log(LogLevel.ERROR, f'ERROR REGISTERING LISTENER FOR EVENT QUEUE { queue["name"] }')
+
+  def getProcessLoggerFunction(self, queue=None, job=None):
+    """Logs to STDOUT formatted to help get helpful information
+    
+    Parameters
+    ----------
+    queue : int, optional
+      ID for the queue sending this message, by default 0
+    job : int, optional
+      ID for the job sending this message, by default 0
+    """
+    W = str(self.worker["id"]).zfill(2) # no need to pass as argument, will always be the same
+    Q = str(__.get(queue, 'id', 0)).zfill(3)
+    J = str(__.get(job, 'id', 0)).zfill(7)
+    log_queue_key = __.get(queue, 'name', 'main')
+    log_queue = self.logQueues[log_queue_key]
+    def sendLogToQueue(level, message):
+      D = getUtcDate(datetime.now()).strftime("%Y-%m-%dT%H:%M:%SZ")
+      msgs = message.split('\n')
+      for msg in msgs:
+        log_queue.put({
+          'level': level,
+          'content': f'[D={D}/W={W}/Q={Q}/J={J}] - {msg}'
+        })
+    return sendLogToQueue
+
+  def startLoggingLoop(self):
+    self.log = self.getProcessLoggerFunction()
+    # FN DEFINITION --------------------------
+    def loggingLoop():
+      while not self.stopLogging:
+        # * check mailbox for every queue
+        # * log as many messages from same queue as possible
+        # TODO: May set a limit to this, in message count or timeout for other queues)
+        # ! Send messages to log-upload queue
+        # ! flag as CREATE or UPSERT for retention or real-time visualization only
+        for _, log_queue in self.logQueues.items():
+          while not log_queue.empty():
+            msg = log_queue.get()
+            log(msg['level'], msg['content'])
+      self.log(LogLevel.WARNING, 'Logging loop stopped')
+    # END FN DEFINITION --------------------------
+    # Create Queues for each queue:
+    for queue_name, _ in self.queues.items():
+      self.logQueues[queue_name] = SimpleQueue()
+    self.loggingProcess = self.context.Process(target=loggingLoop)
+    self.loggingProcess.start()
+    self.log(LogLevel.SUCCESS, 'Logging loop started')
+
+  def stopLoggingLoop(self):
+    self.stopLogging = True
+    self.loggingProcess.join()
 
   def eventLoop(self, interval=1.0):
     try:
       currentQueues, errors = self.gql.query(queries.all_queues)
       if (errors):
-        log(LogLevel.ERROR, f'errors: {errors}')
+        self.log(LogLevel.ERROR, f'errors: {errors}')
         return
       self.currentQueues = currentQueues
       self.unlockFinishedJobs()
@@ -528,14 +591,26 @@ class ValiotWorker():
       if self.pollingMode == PollingMode.QUERY:
         self.processJobs()
       else:
-        log(LogLevel.INFO, 'Listening to events...')
+        self.log(LogLevel.INFO, 'Listening to events...')
     except Exception as e:
-      log(LogLevel.ERROR, f"Error in worker's event loop!")
-      log(LogLevel.ERROR, f'e: {e}')
+      self.log(LogLevel.ERROR, f"Error in worker's event loop!")
+      self.log(LogLevel.ERROR, f'e: {e}')
     time.sleep(interval)
+  
+  def getWorker(self):
+    worker, errors = self.gql.query_one(
+      queries.get_worker,
+      {'code': self.worker_code}
+    )
+    if errors:
+      log(LogLevel.ERROR, errors)
+      raise Exception('Worker not found')
+    self.worker = worker
 
   def run(self, interval=1.0):
     self.context = mp.get_context('fork')
+    self.getWorker()
+    self.startLoggingLoop()
     self.abortStaleJobs()
     self.updateAvailableQueues()
     if (self.pollingMode == PollingMode.SUBSCRIPTION):
@@ -545,7 +620,8 @@ class ValiotWorker():
       try:
         self.eventLoop(interval=interval)
       except KeyboardInterrupt:
-        log(LogLevel.WARNING, "\nuser's stop signal, exiting...")
+        self.log(LogLevel.WARNING, "\nuser's stop signal, exiting...")
+        self.stopLoggingLoop()
         self.unlockFinishedJobs(force=True)
         self.gql.close()
         break
